#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>

#define LOG 0

#include "pathcompress.h"

inline void do_nothing(char* fmt, ...) { }

#ifndef Dprintf

#define DEBUG 0

#if DEBUG
#define Dprintf printf
#else
#define Dprintf do_nothing
#endif
#endif

#if 1
#define DDprintf Dprintf
#else
#define DDprintf do_nothing
#endif

#if LOG
FILE* log = NULL;
#endif


int pathcompress_iscompressed(char* path) {
  int rvl = 0;

  if (path == NULL) rvl=0;
  else if (strstr(path, ".arch-ids")!=NULL) rvl=0;
  else if (strstr(path, "{arch}")!=NULL) rvl=1;
  else if ((getenv("DONTPATHCOMPRESSARCHIVES")==NULL) && (strstr(path, "{archives}")!=NULL)) rvl=1;
  else {
    char* p = strstr(path,",,get.");
    if (p != NULL) {
      p = strstr(p, "/");
      if (p != NULL) {
	if ((strncmp(p, "/,,pristine", 11)!=0) &&
	    (strncmp(p, "/,,", 3)==0)) {
	  rvl = 1;
	}
      }
    } else if (strstr(path, ",,import.")!=NULL) rvl=1;
    else if (strstr(path, ",,commit.")!=NULL) rvl=1;
    else if (strstr(path, ",,archive-cache-revision.")!=NULL) rvl=1;
    else if (strstr(path, ",,delta-scratch.")!=NULL) rvl=1;
    else if (strstr(path, ",,pristine.")!=NULL) rvl=1;
  }
#if LOG
  //fprintf(log, "iscompressed(%s)=%d\n", path, rvl);
#endif
  //printf("iscompressed(%s)=%d\n", path, rvl);
  return rvl;
}


#define PROCESS_FIND_DIRNAME 0
#define PROCESS_FIND_DIRID 1
#define PROCESS_CHANGE_FILE 2
#define PROCESS_ADD_DIRID 2
#define PROCESS_ADD_DIRNAME 3
#define PROCESS_REPLACE_DIRNAME 4
#define PROCESS_REMOVE_DIRNAME 5
#define PROCESS_REMOVE_DIRID 6

#define PROCESS_COMPRESS 42
#define PROCESS_UNCOMPRESS 24

#define PATH_LEN 2048
#define LINE_LEN 2048
#define DIR_LEN 1024
#define DIRID_LEN 32

char* pathcompress_process_dirnames(char* wd, int action, char* dirid, char* dirname, char* result) {
  FILE* fp = NULL;
  FILE* fpnew = NULL;
  int found;
  int len;
  char* dir;
  char dirnames[PATH_LEN]="";
  char newdirnames[PATH_LEN]="";
  char line[LINE_LEN]="";
  int newdirid = 0;
  char _dirid[DIRID_LEN]="";
  int id;
  int newlines;

#if LOG
  if (log==NULL) log=fopen("/tmp/tla--cygwin.log", "a+");
#endif

  if (dirid == NULL) {
    dirid = _dirid;
  }

  Dprintf("pathcompress_process_dirnames(wd=%s,a=%d,id=%s,name=%s)\n", wd, action, dirid, dirname);

  len = strlen(wd);
  if (wd[len-1]=='/') {
    sprintf(dirnames, "%s%s", wd, PATHCOMPRESS_DIRNAMES);
    if (action >= PROCESS_CHANGE_FILE) {
      sprintf(newdirnames, "%s%s.new", wd, PATHCOMPRESS_DIRNAMES);
    }
  } else {
    sprintf(dirnames, "%s/%s", wd, PATHCOMPRESS_DIRNAMES);
    if (action >= PROCESS_CHANGE_FILE) {
      sprintf(newdirnames, "%s/%s.new", wd, PATHCOMPRESS_DIRNAMES);
    }
  }

  found = 0;
  if (strlen(dirnames)>240) {
    // this is the case where one tries to access a non-existing path
    Dprintf("=-1\n");
    return NULL;
  }

  newlines = 0;

  if (access(dirnames, R_OK) != 0) {
    // this is the case when one tries to create a new directory
    //printf("fe PANIC: cannot open %s\n", dirnames);
    if (action >= PROCESS_CHANGE_FILE) {
      fpnew = fopen(newdirnames, "w+");
      //Dprintf("fopen(%s)=%p\n", newdirnames, fpnew);
    } else {
      return NULL;
    }
  } else {
    fp = fopen(dirnames, "r");
    DDprintf("fopen(%s)=%p\n", dirnames, fp);
    if (fp == NULL) {
      printf("PANIC: cannot open %s\n", dirnames);
      return NULL;
    } else {
      if (action >= PROCESS_CHANGE_FILE) {
	fpnew = fopen(newdirnames, "w+");
	DDprintf("fopen(%s)=%p\n", newdirnames, fpnew);
      }
      while (fgets(line, sizeof(line), fp) != NULL) {
	len = strlen(line);
	// chomp
	if (line[len-1] == 0x0a) { line[len-1] = '\0'; len--; }
	if (line[len-1] == 0x0d) { line[len-1] = '\0'; len--; }

	dir = line;
	while ((*dir != '\t') && (*dir != '\0')) { dir++; }
	if (*dir != '\t') {
	  printf("PANIC: error in dir format\n");
	  return NULL;
	}
	*dir = '\0';
	dir++;
	//printf("c='%s' u='%s' l='%s' d='%s' ", dirid, dirname, line, dir);
	switch (action) {
	case PROCESS_FIND_DIRNAME:
	  if (strcmp(dirid, line)==0) {
	    if (line[0]!='.') {
	      found = 1;
	      strcpy(result, dir);
	    }
	  }
	  break;
	case PROCESS_FIND_DIRID:
	  if (strcmp(dirname, dir)==0) {
	    if (line[0]!='.') {
	      found = 1;
	      strcpy(result, line);
	    }
	  }
	  break;
	case PROCESS_ADD_DIRNAME:
	  // prevent doubles
	  if (strcmp(dirname, dir)==0) {
	    newdirid=-1;
	  }
	  if (newdirid>=0) {
	    if (sscanf(line, "%d", &id)>0) {
	      if (newdirid == id) {
		newdirid++;
	      } else if (newdirid < id) {
		fprintf(fpnew, "%d\t%s\n", newdirid, dirname);
		newdirid=-1;
		newlines++;
		DDprintf("adding to new:+ %d\t%s\n", newdirid, dirname);
	      }
	    }
	  }
	  break;
	case PROCESS_REPLACE_DIRNAME:
	  if (strcmp(dirid, line)==0) {
	    found = 1;
	    sprintf(dir, dirname);
	  }
	  break;
	case PROCESS_REMOVE_DIRID:
	  if (strcmp(dirid, line)==0) {
	    DDprintf("remove from new:= %s\n", dirid);
	    found = 1;
	    continue;
	  }
	  break;
	case PROCESS_REMOVE_DIRNAME:
	  if (strcmp(dirname, dir)==0) {
	    DDprintf("remove from new:= %s\n", dirname);
	    found = 1;
	    continue;
	  }
	  break;
	}
	if (fpnew != NULL) {
	  dir[-1]='\t';
	  fprintf(fpnew, "%s\n", line);
	  newlines++;
	  DDprintf("adding to new:= %s\n", line);
	}
      }
      fclose(fp);
    }
  }

  if (fpnew != NULL) {
    switch(action) {
    case PROCESS_ADD_DIRNAME:
      if (newdirid < 0) {
	break;
      }
      sprintf(dirid, "%d", newdirid);
      /* fall trough */
    case PROCESS_ADD_DIRID:
      sprintf(line, "%s\t%s", dirid, dirname);
      fprintf(fpnew, "%s\n", line);
      newlines++;
      DDprintf("adding to new:+ %s\n", line);
      break;
    }

    fclose(fpnew);
    unlink(dirnames);

    if (newlines>0) {
      rename(newdirnames, dirnames);
    } else {
#if 0
      printf("NEW DIRNAMES is empty!\n");
#endif
      unlink(newdirnames);
    }
  }

  if (found) {
    Dprintf("=%s\n", result);
    return result;
  } else {
    Dprintf("=NULL\n");
    return NULL;
  }
}
char* pathcompress_find_dirid(char* wd, char* dirname, char* dirid) {
  Dprintf("pathcompress_find_dirid(%s,%s)\n",wd,dirname);
  return pathcompress_process_dirnames(wd, PROCESS_FIND_DIRID, NULL, dirname, dirid);
}
char* pathcompress_find_dirname(char* wd, char* dirid, char* dirname) {
  Dprintf("pathcompress_find_dirname(%s,%s)\n",wd,dirid);
  return pathcompress_process_dirnames(wd, PROCESS_FIND_DIRNAME, dirid, NULL, dirname);
}
void pathcompress_add_dirid(char* wd, char* dirid, char* dirname) {
  Dprintf("pathcompress_add_dirid(%s,%s,%s)\n",wd,dirid,dirname);
  pathcompress_process_dirnames(wd, PROCESS_ADD_DIRID, dirid, dirname, NULL);
}
void pathcompress_add_dirname(char* wd, char* dirname, char* dirid) {
  Dprintf("pathcompress_add_dirname(%s,%s)\n",wd,dirname);
  pathcompress_process_dirnames(wd, PROCESS_ADD_DIRNAME, NULL, dirname, dirid);
}
void pathcompress_remove_dirname(char* wd, char* dirname) {
  Dprintf("pathcompress_remove_dirname(%s,%s)\n",wd,dirname);
  pathcompress_process_dirnames(wd, PROCESS_REMOVE_DIRNAME, NULL, dirname, NULL);
}
void pathcompress_replace_dirname(char* wd, char* dirid, char* dirname) {
  Dprintf("pathcompress_replace_dirname(%s,%s,%s)\n",wd,dirid,dirname);
  pathcompress_process_dirnames(wd, PROCESS_REPLACE_DIRNAME, dirid, dirname, NULL);
}

char* pathcompress_process_path(char* path, int action, char* compressed, char* uncompressed, char* abspath) {
  char cwd[PATH_LEN]="";
  char _abspath[PATH_LEN]="";
  char _compressed[PATH_LEN]="";
  char _uncompressed[PATH_LEN]="";
  char dir[1024]="";
  char dirid[DIRID_LEN]="";
  char dirname[DIR_LEN]="";
  char* d;
  char* p;

  char* c;
  char* u;
  char* a;

  int processing=0;

#if LOG
  if (log==NULL) log=fopen("/tmp/tla--cygwin.log", "a+");
#endif

  if (compressed == NULL) {
    compressed = _compressed;
  }

  if (uncompressed == NULL) {
    uncompressed = _uncompressed;
  }

  if (abspath == NULL) {
    abspath = _abspath;
  }

  a = abspath;
  c = compressed;
  u = uncompressed;

  p = path;
  d = dir;

#if LOG
  fprintf(log, "process(%s)", path);
#endif

  if (
      (p[0]=='.' && p[1]=='/')
      || (p[0]=='.' && p[1]=='.' && p[2]=='/')
      || (p[0]!='/')
  ) {
    getcwd(cwd, sizeof(cwd));
    a += sprintf(abspath, "%s/", cwd);
    if (pathcompress_iscompressed(abspath)) {
      processing = 1;
    }
  }

  while (*p != '\0') {
    if ((p[0] == '.') && (p[1] == '/')) {
      p+=2;
      *u++=*c++='.';
      *u++=*c++='/';
      continue;
    }
    if ((p[0] == '.') && (p[1] == '.') && (p[2] == '/')) {
      p+=3;
      *u++=*c++='.';
      *u++=*c++='.';
      *u++=*c++='/';
      a--;
      while (a[-1] != '/') a--;
      *a = '\0';
      /* TODO: check for .. out of {arch} */
      continue;
    }
    if (*p == '/') {
      if ((u==uncompressed) || (u[-1] != '/')) {
	*u++=*p;
	*c++=*p;
	*a++=*p;
      }
      p++;
    }
    d=dir;
    while ((*p != '/') && (*p != '\0')) {
      *d++=*p++;
    }
    *d = '\0';
    if (d==dir) {
      u += sprintf(u, "/");
      c += sprintf(c, "/");
      a += sprintf(a, "/");
    } else if (!processing) {
      u += sprintf(u, "%s", dir);
      c += sprintf(c, "%s", dir);
      a += sprintf(a, "%s", dir);
      *a = '\0';
      if (pathcompress_iscompressed(abspath)) {
	processing=1;
      }
    } else {
      *a = '\0';
      *c = '\0';
      *u = '\0';
      switch (action) {
      case PROCESS_UNCOMPRESS:
	if (pathcompress_find_dirname(abspath, dir, dirname) != NULL) {
	  //printf("find(%s,%s)=%s\n", compressed, dir, dirname);
	  u += sprintf(u, "%s", dirname);
	} else {
	  u += sprintf(u, "%s", dir);
	}
	c += sprintf(c, "%s", dir);
	a += sprintf(a, "%s", dir);
	break;
      case PROCESS_COMPRESS:
	if (pathcompress_find_dirid(abspath, dir, dirid) != NULL) {
	  //printf("find(%s,%s)=%s\n", abspath, dir, dirid);
	  c += sprintf(c, "%s", dirid);
	  a += sprintf(a, "%s", dirid);
	} else {
	  c += sprintf(c, "%s", dir);
	  a += sprintf(a, "%s", dir);
	}
	u += sprintf(u, "%s", dir);
	break;
      }
      *a = *c =*u = '\0';
      //printf("a=%s u=%s c=%s\n", abspath, uncompressed, compressed);
    }
    *u++=*p;
    *c++=*p;
    *a++=*p;
    if (*p == '\0') {
      break;
    }
    p++;
  }
  *a = *c =*u = '\0';

  switch(action) {
  case PROCESS_COMPRESS:
#if LOG
    fprintf(log, "=%s\n", abspath);
#endif
    if (strlen(abspath)>240) {
      //printf("PANIC: compressed path too long \nc=%s\nu=%s\na=%s\n", compressed, uncompressed, abspath);
      return NULL;
    }
    return compressed;
  case PROCESS_UNCOMPRESS:
#if LOG
    fprintf(log, "=s\n", abspath);
#endif
    return uncompressed;
  default:
#if LOG
    fprintf(log, "=%p\n", NULL);
#endif
    return NULL;
  }
}

char* pathcompress_uncompress_path(char* path, char* uncompressed) {
  Dprintf("pathcompress_uncompress_path(%s)", path);
  return pathcompress_process_path(path, PROCESS_UNCOMPRESS, NULL, uncompressed, NULL);
}

char* pathcompress_uncompress_path_ex(char* path, char* uncompressed, char* abspath) {
  Dprintf("pathcompress_uncompress_path_ex(%s)", path);
  return pathcompress_process_path(path, PROCESS_UNCOMPRESS, NULL, uncompressed, abspath);
}

char* pathcompress_compress_path(char* path, char* compressed) {
  Dprintf("pathcompress_compress_path(%s)", path);
  return pathcompress_process_path(path, PROCESS_COMPRESS, compressed, NULL, NULL);
}

char* pathcompress_compress_path_ex(char* path, char* compressed, char* abspath) {
  Dprintf("pathcompress_compress_path_ex(%s)", path);
  return pathcompress_process_path(path, PROCESS_COMPRESS, compressed, NULL, abspath);
}

int pathcompress_split_path(char* path, char* abspath, char* wd, char* dirname) {
  char compressed[PATH_LEN]="";
  char uncompressed[PATH_LEN]="";
  char _abspath[PATH_LEN]="";
  char dir[DIR_LEN]="";
  char* p;

  if (abspath == NULL) {
    abspath = _abspath;
  }

  strcpy(uncompressed, path);
  p = &uncompressed[strlen(uncompressed)];
  while((p>uncompressed) && (p[-1]!='/')) p--;
  strcpy(dir, p);

  pathcompress_compress_path_ex(uncompressed, compressed, abspath);

  if (abspath != NULL) {
    p = &abspath[strlen(abspath)];
    while((p>abspath) && (p[0]!='/')) p--;
    *p = '\0';
  }


  p = &compressed[strlen(compressed)];
  while((p>compressed) && (p[0]!='/')) p--;
  *p = '\0';


  if (wd != NULL) {
    strcpy(wd, compressed);
  }
  if (dirname != NULL) {
    strcpy(dirname, dir);
  }
  if (!pathcompress_iscompressed(abspath)) {
    return 0;
  }
  return 1;
}


void pathcompress_add_path(char* path, char* abspath) {
  char compressed[PATH_LEN]="";
  char dirid[DIRID_LEN]="";
  char dirname[DIR_LEN]="";
  char* p;
  int rvl;

  rvl = pathcompress_split_path(path, abspath, compressed, dirname);
  if (rvl) {
    Dprintf("pathcompress_add_path(%s - a=%s)\n", path ,abspath);
    pathcompress_add_dirname(compressed, dirname, dirid);
    if (abspath != NULL) {
      p = abspath+strlen(abspath);
      //if (*p == '/') p--;
      //while ((*p != '/') && (p > abspath)) p--;

      if (*p == '/') {
	sprintf(p, "%s", dirid);
      } else {
	sprintf(p, "/%s", dirid);
      }
    }
    Dprintf("=%s\n", dirid);
  }
}

void pathcompress_remove_path(char* path) {
  char compressed[PATH_LEN]="";
  char dirid[DIRID_LEN]="";
  char dirname[DIR_LEN]="";
  int rvl;

  rvl = pathcompress_split_path(path, NULL, compressed, dirname);
#if 0
  if (!rvl) {
    rvl = pathcompress_iscompressed(path);
  }
#endif
  if (rvl) {
    Dprintf("pathcompress_remove_path(%s - %s - %s)\n", path, compressed, dirname);
    pathcompress_remove_dirname(compressed, dirname);
    Dprintf("=%s\n", dirid);
  }
}

void pathcompress_add_pathid(char* path) {
  char compressed[PATH_LEN]="";
  char abspath[PATH_LEN]="";
  char dirname[DIR_LEN]="";
  int rvl;

  rvl = pathcompress_split_path(path, abspath, compressed, dirname);
#if 1
  if (!rvl) {
    rvl = pathcompress_iscompressed(abspath);
  }
#endif
  if (rvl) {
    pathcompress_compress_path(path, compressed);
  }

  if (rvl) {
    Dprintf("pathcompress_add_pathid(%s)\n", path);
    pathcompress_add_dirid(compressed, ".", dirname);
    pathcompress_add_dirid(compressed, "..", abspath);
  }
}

void pathcompress_update_pathid(char* path) {
  char compressed[PATH_LEN]="";
  char abspath[PATH_LEN]="";
  char dirname[DIR_LEN]="";
  int rvl;
  char* p;

  rvl = pathcompress_split_path(path, abspath, compressed, dirname);
#if 1
  if (!rvl) {
    rvl = pathcompress_iscompressed(abspath);
  }
#endif
  if (rvl) {
    pathcompress_compress_path(path, compressed);
    Dprintf("pathcompress_update_pathid(%s - c=%s)\n", path, compressed);
    pathcompress_replace_dirname(compressed, ".", dirname);
    pathcompress_replace_dirname(compressed, "..", abspath);
  }
}
